/*
 * bqdump - bq27410 register dumper
 *
 * Copyright (c) 2017, Gražvydas Ignotas
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <linux/i2c.h>
#include <linux/i2c-dev.h>

#define ADDRESS 0x55

#define ARRAY_SIZE(a) (sizeof(a) / sizeof(a[0]))

static int bq_read(int fd, __u8 reg, int len)
{
	struct i2c_rdwr_ioctl_data rdwr;
	struct i2c_msg msg[2];
	unsigned char data[2];
	int ret;

	assert(len == 1 || len == 2);

	msg[0].addr = ADDRESS;
	msg[0].flags = 0;
	msg[0].buf = &reg;
	msg[0].len = sizeof(reg);
	msg[1].addr = ADDRESS;
	msg[1].flags = I2C_M_RD;
	msg[1].buf = data;
	msg[1].len = len;

	rdwr.msgs = msg;
	rdwr.nmsgs = ARRAY_SIZE(msg);

	// I2C_SMBUS also works, but reads sometimes return 0xff?
	ret = ioctl(fd, I2C_RDWR, &rdwr);
	if (ret != ARRAY_SIZE(msg)) {
		fprintf(stderr, "read at %02x failed: ", reg);
		perror(NULL);
		exit(1);
	}
	ret = data[0];
	if (len > 1)
		ret |= data[1] << 8;
	return ret;
}

static void bq_write(int fd, __u8 reg, int val, int len)
{
	struct i2c_rdwr_ioctl_data rdwr;
	struct i2c_msg msg;
	unsigned char data[3];
	int ret;

	assert(len == 1 || len == 2);

	data[0] = reg;
	data[1] = val;
	data[2] = val >> 8;
	msg.addr = ADDRESS;
	msg.flags = 0;
	msg.buf = data;
	msg.len = 1 + len;

	rdwr.msgs = &msg;
	rdwr.nmsgs = 1;

	ret = ioctl(fd, I2C_RDWR, &rdwr);
	if (ret != 1) {
		fprintf(stderr, "write at %02x failed: ", reg);
		perror(NULL);
		exit(1);
	}
}

static int dump_control(int fd, int cntl, const char *name)
{
	int ret;

	bq_write(fd, 0, cntl, 2);
	ret = bq_read(fd, 0x00, 2);
	printf("%-16s%04x\n", name, ret);
	return ret;
}

static void dump_bits(const char * const *desc, int bits)
{
	int i, len;
	
	for (i = 0; i < 16; i++) {
		if (!desc[i]) {
			if (bits & (1 << i))
				printf(" unknown bit %d set!?\n", i);
			continue;
		}
		len = printf(" %s: ", desc[i]);
		while (len++ < 35)
			putc(' ', stdout);
		printf("%s\n", (bits & (1 << i)) ? "yes" : "no");
	}
}

enum val_type {
	DT_SIMPLE = 0,
	DT_TEMP_K,
	DT_TEMP_C,
	DT_SOH,
	DT_HEX,
};

enum mem_type {
	MT_8,
	MT_16LE,
	MT_16BE,
};

static void read_dump_t(int fd, int reg, const char *name,
	const char *unit_name, enum val_type vtype, enum mem_type mtype)
{
	const char * const soh_types[] = {
		"invalid", "instant", "initial", "learned"
	};
	unsigned int uval;
	int len, val;

	val = bq_read(fd, reg, (mtype == MT_8) ? 1 : 2);
	if (mtype == MT_16BE)
		val = ((val & 0xff) << 8) | (val >> 8);

	len = printf("%s: ", name);
	while (len++ < 35)
		putc(' ', stdout);
	printf(mtype == MT_8 ? "  %02x" : "%04x", val & 0xffff);

	switch (vtype) {
	case DT_TEMP_K:
		val -= 2731;
		// fallthrough
	case DT_TEMP_C:
		val = (signed short)val;
		printf(" %5.1fC\n", val / 10.0);
		break;
	case DT_SOH:
		uval = val >> 8;
		printf(" %5d%% (%s)\n", val & 0xff,
			uval < ARRAY_SIZE(soh_types) ?
				soh_types[uval] : "borked");
		break;
	case DT_HEX:
		printf("\n");
		break;
	default:
		printf(" %5d%s\n", (signed short)val, unit_name);
		break;
	}
}

static void read_dump(int fd, int reg, const char *name,
	const char *unit_name, enum val_type type)
{
	read_dump_t(fd, reg, name, unit_name, type, MT_16LE);
}

int main(int argc, char *argv[])
{
	int ret, fd;

	fd = open("/dev/i2c-1", O_RDWR);
	if (fd < 0) {
		perror("i2c open");
		return 1;
	}

	bq_write(fd, 0, 0, 2);
	ret = bq_read(fd, 0x00, 2);
	printf("CONTROL_STATUS: %04x\n", ret);
	static const char *n_status[] = {
		NULL,
		"voltages ok for Qmax",
		"Ra table updates are disabled",
		"constant-power model",
		"in SLEEP mode",
		NULL,
		"entry into HIBERNATE",
		"initialization is complete",
		"resistance has been updated", // 8
		"Qmax has updated",
		"board calibration routine",
		"Coulomb Counter Auto-Calibration",
		"in calibration mode",
		"SEALED",
		"Watchdog Reset",
		"enabled for SHUTDOWN"
	};
	dump_bits(n_status, ret);
	dump_control(fd, 0x01, "DEVICE_TYPE:");
	dump_control(fd, 0x02, "FW_VERSION:");
	dump_control(fd, 0x04, "DM_CODE:");
	dump_control(fd, 0x08, "CHEM_ID:");
	ret = bq_read(fd, 0x06, 2);
	printf("Flags:          %04x\n", ret);
	static const char * const n_flags[] = {
		"Discharging detected",
		"StateOfCharge() <= SOCF",
		"StateOfCharge() <= SOC1",
		"Battery insertion detected",
		"in CONFIG UPDATE mode",
		"POR or RESET occurred",
		NULL,
		"OCV measurement is performed",
		"Fast charging allowed", // 8
		"Full charge is detected",
		NULL, NULL, NULL, NULL,
		"Under-Temperature condition",
		"Over-Temperature condition"
	};
	dump_bits(n_flags, ret);
	read_dump(fd, 0x02, "Temperature", "", DT_TEMP_K);
	read_dump(fd, 0x04, "Voltage", "mV", 0);
	read_dump(fd, 0x08, "NominalAvailableCapacity", "mAh", 0);
	read_dump(fd, 0x0a, "FullAvailableCapacity", "mAh", 0);
	read_dump(fd, 0x0c, "RemainingCapacity", "mAh", 0);
	read_dump(fd, 0x0e, "FullChargeCapacity", "mAh", 0);
	read_dump(fd, 0x10, "AverageCurrent", "mA", 0);
	read_dump(fd, 0x12, "StandbyCurrent", "mA", 0);
	read_dump(fd, 0x14, "MaxLoadCurrent", "mA", 0);
	read_dump(fd, 0x16, "Undocumented16", "mAh?", 0);
	read_dump(fd, 0x18, "AveragePower", "mW", 0);
	read_dump(fd, 0x1a, "Undocumented1a", "", 0);
	read_dump(fd, 0x1c, "StateOfCharge", "%", 0);
	read_dump(fd, 0x1e, "InternalTemperature", "", DT_TEMP_K);
	read_dump(fd, 0x20, "StateOfHealth", "", DT_SOH);
	read_dump(fd, 0x22, "Undocumented22", "", 0);
	read_dump(fd, 0x22, "Undocumented22", "", 0);
	read_dump(fd, 0x24, "Undocumented24", "", 0);
	read_dump(fd, 0x26, "Undocumented26", "", 0);
	read_dump(fd, 0x28, "RemainingCapacityUnfiltered", "mAh", 0);
	read_dump(fd, 0x2a, "RemainingCapacityFiltered", "mAh", 0);
	read_dump(fd, 0x2c, "FullChargeCapacityUnfiltered", "mAh", 0);
	read_dump(fd, 0x2e, "FullChargeCapacityFiltered", "mAh", 0);
	read_dump(fd, 0x30, "StateOfChargeUnfiltered", "%", 0);
	// 32-39 - EREMOTEIO
	read_dump(fd, 0x3a, "OpConfig", "", DT_HEX);
	read_dump(fd, 0x3c, "DesignCapacity", "mAh", 0);
	// 3e-61 - memory IO
	// 62-63,65 - EREMOTEIO
	read_dump_t(fd, 0x64, "Undocumented64", "", 0, MT_8);
	read_dump(fd, 0x66, "Undocumented66", "", 0);
	read_dump(fd, 0x68, "Undocumented68", "", 0);
	read_dump(fd, 0x6a, "Undocumented6a", "", 0);
	read_dump(fd, 0x6c, "Undocumented6c", "", 0);
	read_dump(fd, 0x6e, "Undocumented6e", "", 0);
	// 70-79 - EREMOTEIO
	read_dump(fd, 0x7a, "Undocumented7a", "", 0);
	read_dump(fd, 0x7c, "Undocumented7c", "", 0);
	read_dump(fd, 0x7e, "Undocumented7e", "", 0);
	// 80-ff - mirror of 00-7f

	printf("\nData Memory:\n");
	static const struct {
		unsigned char class;
		unsigned char offset;
		enum val_type vtype;
		enum mem_type mtype;
		const char *name;
		const char *unit;
	} mdesc[] = {
		{   2,  0, DT_TEMP_C, MT_16BE, "Over Temp", "" },
		{   2,  2, DT_TEMP_C, MT_16BE, "Under Temp", "" },
		{   2,  4, DT_TEMP_C, MT_8,    "Temp Hys", "" },
		{  36,  3, 0,         MT_8,    "TCA Set", "%" },
		{  36,  4, 0,         MT_8,    "TCA Clear", "%" },
		// ...
		{  82,  2, DT_HEX,    MT_8,    "Update Status", "" },
		{  82,  3, 0,         MT_16BE, "Reserve Cap", "mAh" },
		{  82,  5, DT_HEX,    MT_8,    "Load Select/Mode", "" },
		{  82,  6, 0,         MT_16BE, "Q Invalid MaxV", "mV" },
		{  82,  8, 0,         MT_16BE, "Q Invalid MinV", "mV" },
		{  82, 10, 0,         MT_16BE, "Design Capacity", "mAh" },
		{  82, 12, 0,         MT_16BE, "Design Energy", "mWh" },
		{  82, 14, 0,         MT_16BE, "Default Design Cap", "mAh" },
		{  82, 16, 0,         MT_16BE, "Terminate Voltage", "mV" },
		// ...
		{  82, 27, 0,         MT_16BE, "Taper Rate", "Hr/10" },
		{  82, 29, 0,         MT_16BE, "Taper Voltage", "mV" },
		{  82, 31, 0,         MT_16BE, "Sleep Current", "mA" },
		{  82, 33, 0,         MT_16BE, "V at Chg Term", "mV" },
	};
	int i, class = -1, block = -1;

	bq_write(fd, 0x61, 0, 1); // BlockDataControl()

	for (i = 0; i < ARRAY_SIZE(mdesc); i++) {
		if (mdesc[i].class != class
		    || mdesc[i].offset / 32 != block)
		{
			class = mdesc[i].class;
			block = mdesc[i].offset / 32;
			bq_write(fd, 0x3e, class, 1); // DataClass()
			bq_write(fd, 0x3f, block, 1); // DataBlock()
			usleep(1000);
		}
		read_dump_t(fd, 0x40 + mdesc[i].offset % 32,
			mdesc[i].name, mdesc[i].unit, mdesc[i].vtype,
			mdesc[i].mtype);
	}

	close(fd);

	return 0;
}
